"""
The DryRunDeployManager implements the DeployManager interface but does not
actually interact with the cluster and instead holds the state of the cluster in
a local map.
"""

# Standard
from datetime import datetime, timedelta
from functools import partial
from queue import Empty, Queue
from threading import RLock
from typing import Callable, Iterator, List, Optional, Tuple
import copy
import operator
import random
import uuid

# First Party
import alog

# Local
from ..managed_object import ManagedObject
from ..utils import merge_configs
from .base import DeployManagerBase, DeployMethod
from .kube_event import KubeEventType, KubeWatchEvent
from .owner_references import update_owner_references

log = alog.use_channel("DRY-RUN")

# Lock to ensure disable/deploys are thread safe
DRY_RUN_CLUSTER_LOCK = RLock()


class DryRunDeployManager(DeployManagerBase):
    """
    Deploy manager which doesn't actually deploy!
    """

    def __init__(
        self,
        resources=None,
        owner_cr=None,
        strict_resource_version=False,
        generate_resource_version=True,
    ):
        """Construct with a static value to use for whether or not the functions
        should report change.
        """
        self._owner_cr = owner_cr
        self._cluster_content = {}
        self.strict_resource_version = strict_resource_version
        self.generate_resource_version = generate_resource_version

        # Dicts of registered watches and watchers
        self._watches = {}
        self._finalizers = {}

        # Deploy provided resources
        self._deploy(resources or [], call_watches=False, manage_owner_references=False)

    ## Interface ###############################################################

    def deploy(
        self,
        resource_definitions,
        manage_owner_references=True,
        method: DeployMethod = DeployMethod.DEFAULT,
        **_,
    ):
        log.info("DRY RUN deploy")
        return self._deploy(
            resource_definitions,
            manage_owner_references=manage_owner_references,
            method=method,
        )

    def disable(self, resource_definitions):
        log.info("DRY RUN disable")
        changed = False
        for resource in resource_definitions:
            api_version = resource.get("apiVersion")
            kind = resource.get("kind")
            name = resource.get("metadata", {}).get("name")
            namespace = resource.get("metadata", {}).get("namespace")
            _, content = self.get_object_current_state(
                kind=kind, api_version=api_version, namespace=namespace, name=name
            )
            if content is not None:
                changed = True

                # Set resource finalizers
                with DRY_RUN_CLUSTER_LOCK:
                    self._cluster_content[namespace][kind][api_version][name][
                        "metadata"
                    ]["deletionTimestamp"] = datetime.now().strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                    self._cluster_content[namespace][kind][api_version][name][
                        "metadata"
                    ]["deletionGracePeriodSeconds"] = 0

                # Call any registered finalizers
                for key, callback in self._get_registered_watches(
                    api_version, kind, namespace, name, finalizer=True
                ):
                    log.debug2(
                        "Calling registered finalizer [%s] for [%s]", callback, key
                    )
                    callback(self._cluster_content[namespace][kind][api_version][name])

                # If finalizers have been cleared and object hasn't already been deleted then
                # remove the key
                current_obj = (
                    self._cluster_content.get(namespace, {})
                    .get(kind, {})
                    .get(api_version, {})
                    .get(name, {})
                )
                if current_obj and not current_obj.get("metadata", {}).get(
                    "finalizers", []
                ):
                    with DRY_RUN_CLUSTER_LOCK:
                        self._delete_key(namespace, kind, api_version, name)

        return True, changed

    def get_object_current_state(self, kind, name, namespace=None, api_version=None):
        log.info(
            "DRY RUN get_object_current_state of [%s/%s] in [%s]", kind, name, namespace
        )

        # Look in the cluster state
        matches = []
        kind_entries = self._cluster_content.get(namespace, {}).get(kind, {})
        log.debug3("Kind entries: %s", kind_entries)
        for api_ver, entries in kind_entries.items():
            log.debug3("Checking api_version [%s // %s]", api_ver, api_version)
            if name in entries and (api_ver == api_version or api_version is None):
                matches.append(entries[name])
        log.debug(
            "Found %d matches for [%s/%s] in %s", len(matches), kind, name, namespace
        )
        if len(matches) == 1:
            return True, copy.deepcopy(matches[0])
        return True, None

    def filter_objects_current_state(
        self,
        kind,
        namespace=None,
        api_version=None,
        label_selector=None,
        field_selector=None,
    ):  # pylint: disable=too-many-arguments
        log.info(
            "DRY RUN filter_objects_current_state of [%s] in [%s]", kind, namespace
        )
        # Look in the cluster state
        matches = []
        kind_entries = self._cluster_content.get(namespace, {}).get(kind, {})
        log.debug3("Kind entries: %s", kind_entries)
        for api_ver, entries in kind_entries.items():
            # Make sure api version matches
            log.debug3("Checking api_version [%s // %s]", api_ver, api_version)
            if api_ver != api_version and api_version is not None:
                continue

            for resource in entries.values():
                # Make sure Labels Match
                log.debug3("Resource: %s", resource)

                labels = resource.get("metadata", {}).get("labels", {})
                log.debug3("Checking label_selector [%s // %s]", labels, label_selector)
                if label_selector is not None and not _match_selector(
                    labels, label_selector
                ):
                    continue

                # Only do the work for field selector if one exists
                log.debug3("Checking field_selector [%s]", field_selector)
                if field_selector is not None and not _match_selector(
                    _convert_dict_to_dot(resource),
                    field_selector,
                ):
                    continue

                # Add deep copy of entry to matches list
                matches.append(copy.deepcopy(resource))

        return True, matches

    def set_status(
        self,
        kind,
        name,
        namespace,
        status,
        api_version=None,
    ):  # pylint: disable=too-many-arguments
        log.info(
            "DRY RUN set_status of [%s.%s/%s] in %s: %s",
            api_version,
            kind,
            name,
            namespace,
            status,
        )
        object_content = self.get_object_current_state(
            kind, name, namespace, api_version
        )[1]
        if object_content is None:
            log.debug("Did not find [%s/%s] in %s", kind, name, namespace)
            return False, False
        prev_status = object_content.get("status")
        object_content["status"] = status
        self._deploy([object_content], call_watches=False)
        return True, prev_status != status

    def watch_objects(  # pylint: disable=too-many-arguments,too-many-locals,unused-argument
        self,
        kind: str,
        api_version: Optional[str] = None,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        resource_version: Optional[str] = None,
        timeout: Optional[int] = 15,
        **kwargs,
    ) -> Iterator[KubeWatchEvent]:
        """Watch the DryRunDeployManager for resource changes by registering
        callbacks"""

        event_queue = Queue()
        resource_map = {}

        def add_event(resource_map: dict, manifest: dict):
            """Callback triggered when resources are deployed"""
            resource = ManagedObject(manifest)
            event_type = KubeEventType.ADDED

            watch_key = self._watch_key(
                api_version=resource.api_version,
                kind=resource.kind,
                namespace=resource.namespace,
                name=resource.name,
            )
            if watch_key in resource_map:
                log.debug4("Watch key detected, setting Modified event type")
                event_type = KubeEventType.MODIFIED

            resource_map[watch_key] = resource
            event = KubeWatchEvent(
                type=event_type,
                resource=resource,
            )
            event_queue.put(event)

        def delete_event(resource_map: dict, manifest: dict):
            """Callback triggered when resources are disabled"""
            resource = ManagedObject(manifest)
            watch_key = self._watch_key(
                api_version=resource.api_version,
                kind=resource.kind,
                namespace=resource.namespace,
                name=resource.name,
            )
            if watch_key in resource_map:
                del resource_map[watch_key]

            event = KubeWatchEvent(
                type=KubeEventType.DELETED,
                resource=resource,
            )
            event_queue.put(event)

        # Get initial resources
        _, manifests = self.filter_objects_current_state(
            kind=kind,
            api_version=api_version,
            namespace=namespace,
            label_selector=label_selector,
            field_selector=field_selector,
        )
        for manifest in manifests:
            resource = ManagedObject(manifest)
            watch_key = self._watch_key(
                kind=resource.kind,
                api_version=resource.api_version,
                name=resource.name,
                namespace=resource.namespace,
            )
            resource_map[watch_key] = resource

            event = KubeWatchEvent(type=KubeEventType.ADDED, resource=resource)
            log.debug2("Yielding initial event %s", event)
            yield event

        end_time = datetime.max
        if timeout:
            end_time = datetime.now() + timedelta(seconds=timeout)

        # Register callbacks
        self.register_watch(
            api_version=api_version,
            kind=kind,
            namespace=namespace,
            name=name,
            callback=partial(add_event, resource_map),
        )
        self.register_finalizer(
            api_version=api_version,
            kind=kind,
            namespace=namespace,
            name=name,
            callback=partial(delete_event, resource_map),
        )

        # Yield any events from the callback queue
        log.debug2("Waiting till %s", end_time)
        while True:
            sec_till_end = (end_time - datetime.now()).seconds or 1
            try:
                event = event_queue.get(timeout=sec_till_end)
                log.debug2("Yielding event %s", event)
                yield event
            except Empty:
                pass

            if datetime.now() > end_time:
                return

    ## Dry Run Methods #########################################################
    def register_watch(  # pylint: disable=too-many-arguments
        self,
        api_version: str,
        kind: str,
        callback: Callable[[dict], None],
        namespace="",
        name="",
    ):
        """Register a callback to watch for deploy events on a given
        api_version/kind
        """
        watch_key = self._watch_key(
            api_version=api_version, kind=kind, namespace=namespace, name=name
        )
        log.debug("Registering watch for %s", watch_key)
        self._watches.setdefault(watch_key, []).append(callback)

    def register_finalizer(  # pylint: disable=too-many-arguments
        self,
        api_version: str,
        kind: str,
        callback: Callable[[dict], None],
        namespace="",
        name="",
    ):
        """Register a callback to call on deletion events on a given
        api_version/kind
        """
        watch_key = self._watch_key(
            api_version=api_version, kind=kind, namespace=namespace, name=name
        )
        log.debug("Registering finalizer for %s", watch_key)
        self._finalizers.setdefault(watch_key, []).append(callback)

    ## Implementation Details ##################################################

    @staticmethod
    def _watch_key(api_version="", kind="", namespace="", name=""):
        return ":".join([api_version or "", kind or "", namespace or "", name or ""])

    def _get_registered_watches(  # pylint: disable=too-many-arguments
        self,
        api_version: str = "",
        kind: str = "",
        namespace: str = "",
        name: str = "",
        finalizer: bool = False,
    ) -> List[Tuple[str, Callable]]:
        # Get the scoped watch key
        resource_watch_key = self._watch_key(
            api_version=api_version, kind=kind, namespace=namespace, name=name
        )
        namespaced_watch_key = self._watch_key(
            api_version=api_version, kind=kind, namespace=namespace
        )
        global_watch_key = self._watch_key(api_version=api_version, kind=kind)

        # Get which watch list we're pulling from
        callback_map = self._watches
        if finalizer:
            callback_map = self._finalizers

        output_list = []
        log.debug3(
            "Looking for resourced key: %s namespace key %s global key %s",
            resource_watch_key,
            namespaced_watch_key,
            global_watch_key,
        )
        for key, callback_list in callback_map.items():
            if key in [resource_watch_key, namespaced_watch_key, global_watch_key]:
                log.debug3("%d Callbacks found for key %s", len(callback_list), key)
                for callback in callback_list:
                    output_list.append((key, callback))

        return output_list

    def _delete_key(self, namespace, kind, api_version, name):
        del self._cluster_content[namespace][kind][api_version][name]
        if not self._cluster_content[namespace][kind][api_version]:
            del self._cluster_content[namespace][kind][api_version]
        if not self._cluster_content[namespace][kind]:
            del self._cluster_content[namespace][kind]
        if not self._cluster_content[namespace]:
            del self._cluster_content[namespace]

    def _deploy(
        self,
        resource_definitions,
        call_watches=True,
        manage_owner_references=True,
        method: DeployMethod = DeployMethod.DEFAULT,
    ):
        log.info("DRY RUN deploy")
        changes = False
        for resource in resource_definitions:
            api_version = resource.get("apiVersion")
            kind = resource.get("kind")
            name = resource.get("metadata", {}).get("name")
            namespace = resource.get("metadata", {}).get("namespace")
            log.debug(
                "DRY RUN deploy [%s/%s/%s/%s]", namespace, kind, api_version, name
            )
            log.debug4(resource)

            # If owner CR configured, add ownerReferences
            if self._owner_cr and manage_owner_references:
                log.debug2("Adding dry-run owner references")
                update_owner_references(self, self._owner_cr, resource)
                log.debug3(
                    "All owner references: %s", resource["metadata"]["ownerReferences"]
                )

            with DRY_RUN_CLUSTER_LOCK:
                entries = (
                    self._cluster_content.setdefault(namespace, {})
                    .setdefault(kind, {})
                    .setdefault(api_version, {})
                )
                current = copy.deepcopy(entries.get(name, {}))
                old_resource_version = current.get("metadata", {}).pop(
                    "resourceVersion", None
                )
                changes = changes or (current != resource)

                if "metadata" not in resource:
                    resource["metadata"] = {}

                if (
                    self.strict_resource_version
                    and resource["metadata"].get("resourceVersion")
                    and old_resource_version
                    and resource["metadata"].get("resourceVersion")
                    != old_resource_version
                ):
                    log.warning(
                        "Unable to deploy resource. resourceVersion is out of date"
                    )
                    return False, False

                resource["metadata"]["creationTimestamp"] = entries.get(
                    "metadata", {}
                ).get("creationTimestamp", datetime.now().isoformat())
                resource["metadata"]["uid"] = entries.get("metadata", {}).get(
                    "uid", str(uuid.uuid4())
                )

                if self.generate_resource_version:
                    resource["metadata"]["resourceVersion"] = str(
                        random.randint(1, 1000)
                    ).zfill(5)

                # Depending on the deploy method either update or fully replace the object
                if method == DeployMethod.DEFAULT or method == DeployMethod.REPLACE:
                    entries[name] = resource
                else:
                    if name in entries:
                        entries[name] = merge_configs(entries[name], resource)
                    # If the object doesn't already exist then just add it
                    else:
                        entries[name] = resource

            # Call any registered watches
            if call_watches:
                for key, callback in self._get_registered_watches(
                    api_version, kind, namespace, name
                ):
                    log.debug2("Calling registered watch [%s] for [%s]", callback, key)
                    callback(resource)

            # Delete Key if it has already been disabled and doesn't have finalizers
            if self._cluster_content[namespace][kind][api_version][name].get(
                "metadata", {}
            ).get("deletionTimestamp") and not self._cluster_content[namespace][kind][
                api_version
            ][
                name
            ].get(
                "metadata", {}
            ).get(
                "finalizers"
            ):
                with DRY_RUN_CLUSTER_LOCK:
                    self._delete_key(namespace, kind, api_version, name)

        return True, changes


def _match_selector(values, value_selector) -> bool:  # pylint: disable=too-many-locals
    """This function implements the kubernetes selector to determine if
    a set of values matches the selector. For the complete documentation regardin
    selectors see:
    https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set
    """
    log.debug2("DRY RUN match_selector [%s/%s]", values, value_selector)

    # Configure List of Operators used in splitting
    equality_ops = ["=", "==", "!="]
    # Note the spaces are required to correctly distinguish between operators and random characters
    set_ops = [" in ", " notin "]
    existence_ops = ["!", ""]

    # Define dict of ops to python functions. Tried to use operator library as much as possible but
    # four custom functions were needed
    def _in(a, b):  # pylint: disable=invalid-name
        return a in b

    def not_in(a, b):  # pylint: disable=invalid-name
        return not _in(a, b)

    def exists(a, _):  # pylint: disable=invalid-name
        return a is not None

    def not_exists(a, _):  # pylint: disable=invalid-name
        return a is None

    operator_actions = {
        "=": operator.eq,
        "==": operator.eq,
        "!=": operator.ne,
        " in ": _in,
        " notin ": not_in,
        "!": not_exists,
        "": exists,
    }

    # Sort the list of operators reverse by size so when splitting it doesn't mess up
    # e.g. check splitting by != before = as = is apart of !=.
    operator_list = sorted(operator_actions.keys(), key=len, reverse=True)

    # Loop through all selectors to validate each one
    for selector in _split_selectors(value_selector):
        # Variables set during splitting and used for verification
        action = None
        expected_key = None
        expected_value = None

        # Get the operator in use by the selector. This is done
        # by trying to split by each operator until one works
        for op in operator_list:  # pylint: disable=invalid-name
            # Either split the selector or remove the NonExistence operator
            split_selector = []
            if op in existence_ops:
                split_selector = [selector.replace(op, "")]
            else:
                split_selector = selector.split(op)

            # If Selector doesn't split correctly for equality_ops/set_ops
            if (op in equality_ops or op in set_ops) and len(split_selector) != 2:
                continue

            # If the operator is ! then make sure ! was actually removed from the selector
            if (op == "!") and "!" not in selector:
                continue

            # Get the action and key
            action = operator_actions[op]
            expected_key = split_selector[0].strip()

            # If op is equality then the entire second half is the expected value
            if op in equality_ops:
                expected_value = split_selector[1].strip()

            # If op is set then format the expected values that are in the form (value,value,etc)
            # Into a list
            elif op in set_ops:
                string_value = split_selector[1]
                string_value = string_value.replace("(", "")
                string_value = string_value.replace(")", "")
                expected_value = string_value.split(",")

            # Once the selector is found break out of the loop
            break

        # Get the value at the expected key
        value = values.get(expected_key)

        # Convert all values to strings unless None
        value = str(value).strip() if value is not None else value

        if not action(value, expected_value):
            log.debug3(
                "Label with key: %s and value: %s does not match selector %s",
                expected_key,
                value,
                selector,
            )
            return False

    # If all selectors matched then return True
    return True


def _split_selectors(selector=""):
    """Split up selectors by , but ignoring those surrounded by () e.g.
    'app,app in (frontend, backend)' becomes ['app','app in (frontend, backend)']
    """
    output_list = []
    current_selector = ""
    in_paren = False

    # Loop through selector character by character
    for char in selector:
        # If we've reached a comma not while in parentheses then append to output
        # and start anew
        if char == "," and not in_paren:
            output_list.append(current_selector)
            current_selector = ""
            continue

        # If detect paren then update in_paren
        if char == "(" and not in_paren:
            in_paren = True
        elif char == ")" and in_paren:
            in_paren = False

        # Append the current char
        current_selector += char

    # If current selector is not empty then add it to the output list
    if current_selector:
        output_list.append(current_selector)

    return output_list


def _convert_dict_to_dot(dictionary, prefix=""):
    """Helper function to convert a dictionary to a map
    of strings dotted together. For example {a:{b:1},c:2}
    becomes {a.b:1,c:2}
    """
    if not isinstance(dictionary, dict):
        return {prefix: dictionary}

    output_dict = {}
    for key in dictionary:
        new_key = key if prefix == "" else f"{prefix}.{key}"
        output_dict = {**output_dict, **_convert_dict_to_dot(dictionary[key], new_key)}
    return output_dict



"""
The implementation of the cluster resource interactions to manage a
TemporaryPatch resource type
"""

# Standard
from datetime import datetime
import copy
import json

# First Party
import alog

# Local
from ..component import Component
from ..constants import TEMPORARY_PATCHES_ANNOTATION_NAME
from ..decorator import component
from ..exceptions import ConfigError, assert_cluster, assert_config, assert_precondition
from ..session import Session

log = alog.use_channel("PATCH")


@component("temporarypatch")
class TemporaryPatchComponent(Component):
    """
    The TemporaryPatchComponent is responsible for applying the temporary patch
    annotation to the target oper8 resource type.

    See the ADR for full details:
        docs/adr/03-patches.md
    """

    def __init__(
        self,
        session: Session,
        disabled: bool,
        patch_name: str,
        target_api_version: str,
        target_kind: str,
        target_name: str,
    ):
        """At setup, this component will attempt to fetch the state of the
        target resource and fail out if not possible.

        Args:
            session:  Session
                The session that this Component belongs to
            disabled:  bool
                True if this patch should be removed from the target, False if
                it should be added to the target
            patch_name:  str
                The name of the patch itself that will be used to uniquely
                identify this patch within the target
            target_api_version:  str
                The group/version string for the apiVersion of the target type
            target_kind:  str
                The kind string for the target type
            target_name:  str
                The metadata.name string for the target instance
        """
        super().__init__(session=session, disabled=disabled)

        # Hold the target info for later
        self._patch_name = patch_name
        self._target_api_version = target_api_version
        self._target_kind = target_kind
        self._target_name = target_name

        # Look up the target state
        log.debug2(
            "Looking up target [%s/%s/%s/%s]",
            session.namespace,
            target_api_version,
            target_kind,
            target_name,
        )
        success, target_state = session.deploy_manager.get_object_current_state(
            api_version=target_api_version,
            kind=target_kind,
            name=target_name,
            namespace=session.namespace,
        )
        assert_cluster(
            success,
            "Failed to look up target resource: {}/{}/{}/{}".format(
                session.namespace,
                target_api_version,
                target_kind,
                target_name,
            ),
        )
        assert_precondition(
            disabled or target_state is not None,
            "Could not find target resource: {}/{}/{}/{}".format(
                session.namespace,
                target_api_version,
                target_kind,
                target_name,
            ),
        )
        self._target_state = target_state

    def deploy(self, session: Session) -> bool:
        """The TemporaryPatchComponent runs a custom deploy operation to update
        the target resource with the patch annotation.
        """
        return self._manage_patch(session, remove_patch=False)

    def disable(self, session: Session) -> bool:
        """The TemporaryPatchComponent runs a custom disable operation to remove
        the patch from the target resource.
        """
        if self._target_state is not None:
            return self._manage_patch(session, remove_patch=True)
        return True

    ## Implementation Details ##################################################

    def _manage_patch(self, session: Session, remove_patch: bool) -> bool:
        """Manage the presence of the patch on the target"""
        log.debug2("Patch will be [%s]", "removed" if remove_patch else "added")

        # Look for existing annotation entries on the resource
        log.debug4("Attempting to parse target state: %s", self._target_state)
        annotations = self._target_state["metadata"].get("annotations", {})
        anno_content = annotations.get(TEMPORARY_PATCHES_ANNOTATION_NAME, "{}")
        try:
            existing_patches = json.loads(anno_content)
        except (TypeError, json.decoder.JSONDecodeError) as err:
            raise ConfigError(
                f"Patch annotation is not valid json: {anno_content}"
            ) from err
        log.debug3("Existing Patches: %s", existing_patches)
        assert_config(
            isinstance(existing_patches, dict),
            f"Existing patches formatted incorrectly: {anno_content}",
        )

        # Perform the add/delete operation
        updated_patches = copy.deepcopy(existing_patches)
        if remove_patch:
            log.debug2("Removing patch [%s]", self._patch_name)
            updated_patches.pop(self._patch_name, None)
        elif self._patch_name not in updated_patches:
            log.debug("Adding patch [%s]", self._patch_name)
            # Add this patch to the existing patches and timestamp it. The
            # structure of the patch body holds the api_version and kind for the
            # patch itself so that the Controller for the target resource can
            # look it up without hard-coding either value.
            log.debug2("Adding patch [%s]", self._patch_name)
            updated_patches[self._patch_name] = {
                "timestamp": datetime.now().isoformat(),
                "api_version": session.api_version,
                "kind": session.kind,
            }

        if updated_patches != existing_patches:
            log.debug(
                "Re-applying [%s/%s/%s/%s]",
                session.namespace,
                self._target_api_version,
                self._target_kind,
                self._target_name,
            )

            # Re-serialize the annotations
            self._target_state["metadata"].setdefault("annotations", {})[
                TEMPORARY_PATCHES_ANNOTATION_NAME
            ] = json.dumps(updated_patches)

            # Update the resource in the cluster
            success, _ = session.deploy_manager.deploy(
                resource_definitions=[self._target_state],
                manage_owner_references=False,
            )
            assert_cluster(
                success,
                "Failed to re-apply [{}/{}/{}/{}]".format(
                    session.namespace,
                    self._target_api_version,
                    self._target_kind,
                    self._target_name,
                ),
            )

        # To keep with the deploy API, return success at this point
        return True



"""
Make sure that redis interfaces import cleanly
"""

# Third Party
import pytest

# Local
from oper8.x.datastores.redis.interfaces import IRedisComponent


def test_is_abstract():
    with pytest.raises(TypeError):
        IRedisComponent()



"""
Tests for the AnnotationLeadershipManager
"""
# Standard
from datetime import datetime, timedelta
import threading
import time

# Third Party
import pytest

# Local
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.managed_object import ManagedObject
from oper8.test_helpers.helpers import library_config
from oper8.test_helpers.pwm_helpers import make_managed_object
from oper8.watch_manager.python_watch_manager.leader_election.annotation import (
    AnnotationLeadershipManager,
)

## Helpers #####################################################################


@pytest.mark.timeout(5)
def test_annotation_happy_path():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([resource.definition])
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "duration": "10s",
                "namespace": resource.namespace,
            }
        },
    ):
        lm = AnnotationLeadershipManager(deploy_manager=dm)
        assert lm.acquire()
        assert lm.is_leader()

        child_resource = make_managed_object()
        dm.deploy([child_resource.definition])
        assert lm.acquire_resource(child_resource)
        _, updated_object = dm.get_object_current_state(
            child_resource.kind,
            child_resource.name,
            child_resource.namespace,
            child_resource.api_version,
        )
        assert lm.is_leader(ManagedObject(updated_object))
        assert (
            updated_object.get("metadata", {})
            .get("annotations", {})
            .get("oper8.org/lease-name")
            == resource.name
        )

        assert lm.release_resource(ManagedObject(updated_object))
        _, updated_object = dm.get_object_current_state(
            child_resource.kind,
            child_resource.name,
            child_resource.namespace,
            child_resource.api_version,
        )
        assert (
            not updated_object.get("metadata", {})
            .get("annotations", {})
            .get("oper8.org/lease-name")
        )
        assert lm.release()


@pytest.mark.timeout(5)
def test_annotation_already_owner():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([resource.definition])
    old_time = datetime.now()
    existing_object = make_managed_object(
        annotations={
            "oper8.org/lease-name": resource.name,
            "oper8.org/lease-time": old_time.isoformat(),
        }
    )
    dm.deploy([existing_object.definition])
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "duration": "10s",
                "namespace": resource.namespace,
            }
        },
    ):
        lm = AnnotationLeadershipManager(deploy_manager=dm)
        assert lm.acquire()
        assert lm.is_leader()
        assert lm.acquire_resource(existing_object)
        _, updated_object = dm.get_object_current_state(
            existing_object.kind,
            existing_object.name,
            existing_object.namespace,
            existing_object.api_version,
        )
        assert (
            updated_object.get("metadata", {})
            .get("annotations", {})
            .get("oper8.org/lease-time")
            != old_time.isoformat()
        )
        lm.release_resource(existing_object)


@pytest.mark.timeout(5)
def test_annotation_transition():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([resource.definition])
    old_owner = make_managed_object(kind="Pod", api_version="v1", name="old")
    dm.deploy([old_owner.definition])
    old_time = datetime.now() - timedelta(seconds=30)
    existing_object = make_managed_object(
        annotations={
            "oper8.org/lease-name": old_owner.name,
            "oper8.org/lease-time": old_time.isoformat(),
        }
    )
    dm.deploy([existing_object.definition])
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "duration": "1s",
                "namespace": resource.namespace,
            }
        },
    ):
        lm = AnnotationLeadershipManager(deploy_manager=dm)
        assert lm.acquire()
        assert lm.is_leader()
        assert lm.acquire_resource(existing_object)
        _, updated_lease = dm.get_object_current_state(
            existing_object.kind,
            existing_object.name,
            existing_object.namespace,
            existing_object.api_version,
        )
        assert (
            updated_lease.get("metadata", {})
            .get("annotations", {})
            .get("oper8.org/lease-name")
            == resource.name
        )
        lm.release_resource(existing_object)


@pytest.mark.timeout(5)
def test_annotation_not_owner():
    dm = DryRunDeployManager()
    original_owner = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([original_owner.definition])
    old_time = datetime.now()
    existing_object = make_managed_object(
        annotations={
            "oper8.org/lease-name": original_owner.name,
            "oper8.org/lease-time": old_time.isoformat(),
        }
    )
    dm.deploy([existing_object.definition])
    current_pod = make_managed_object(kind="Pod", api_version="v1", name="current")
    dm.deploy([current_pod.definition])
    with library_config(
        pod_name=current_pod.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "namespace": current_pod.namespace,
                "duration": "60s",
            }
        },
    ):
        lm = AnnotationLeadershipManager(deploy_manager=dm)
        assert not lm.acquire_resource(existing_object)
        _, updated_object = dm.get_object_current_state(
            existing_object.kind,
            existing_object.name,
            existing_object.namespace,
            existing_object.api_version,
        )
        assert (
            updated_object.get("metadata", {})
            .get("annotations", {})
            .get("oper8.org/lease-name")
            == original_owner.name
        )
        lm.release_resource(existing_object)


@pytest.mark.timeout(5)
def test_annotation_competition():
    dm = DryRunDeployManager(strict_resource_version=True)
    first_operator = make_managed_object(kind="Pod", api_version="v1", name="first")
    dm.deploy([first_operator.definition])
    with library_config(
        pod_name=first_operator.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "namespace": first_operator.namespace,
                "duration": "1s",
            }
        },
    ):
        first_manager = AnnotationLeadershipManager(deploy_manager=dm)

    second_operator = make_managed_object(kind="Pod", api_version="v1", name="second")
    dm.deploy([second_operator.definition])
    with library_config(
        pod_name=second_operator.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "namespace": second_operator.namespace,
                "duration": "1s",
            }
        },
    ):
        second_manager = AnnotationLeadershipManager(deploy_manager=dm)

    target_object = make_managed_object()
    dm.deploy([target_object.definition])

    threading.Thread(
        target=first_manager.acquire_resource, args=[target_object]
    ).start()
    threading.Thread(
        target=second_manager.acquire_resource, args=[target_object]
    ).start()

    time.sleep(1)
    # Assert only one of them gained leadership
    _, updated_object = dm.get_object_current_state(
        target_object.kind,
        target_object.name,
        target_object.namespace,
        target_object.api_version,
    )
    assert updated_object.get("metadata", {}).get("annotations", {}).get(
        "oper8.org/lease-name"
    ) in [first_operator.name, second_operator.name]



""" Import All functions, constants, and class from utils module """
# Local
from .common import (
    get_logging_handlers,
    get_operator_namespace,
    get_pod_name,
    obj_to_hash,
    parse_time_delta,
)
from .constants import (
    JOIN_PROCESS_TIMEOUT,
    MIN_SLEEP_TIME,
    RESERVED_PLATFORM_ANNOTATIONS,
    RESOURCE_VERSION_KEEP_COUNT,
    SHUTDOWN_RECONCILE_POLL_TIME,
)
from .log_handler import LogQueueHandler
from .types import (
    ABCSingletonMeta,
    ClassInfo,
    ReconcileProcess,
    ReconcileRequest,
    ReconcileRequestType,
    ResourceId,
    Singleton,
    TimerEvent,
    WatchedResource,
    WatchRequest,
)



"""
Shared constants across the various oper8.x tools
"""

# The key in the CR spec that holds the datastore information
SPEC_DATASTORES = "datastores"

# The key inside an individual spec.<datastore_type> that holds connection info
SPEC_DATASTORE_CONNECTION = "connection"

GLOBAL_SLOT = "global"

DEPS_ANNOTATION = "oper8.org/dependency-hash"

SERVICE_ACCOUNT_NAME_WORKLOAD_SCHEDULER = "workload-scheduler"

PULL_POLICY = "IfNotPresent"



"""This file contains common utilities for detecting if a replace operation is required
for a resource
"""
# Standard
from typing import Any, Callable, List

# Third Party
from openshift.dynamic.apply import recursive_diff, recursive_list_diff

# First Party
import alog

log = alog.use_channel("DMRPLC_UTILS")


def modified_lists(
    current_manifest: dict, desired_manifest: dict, in_list: bool = False
) -> bool:
    """Helper function to check if there are any differences in the lists of the desired manifest.
    This is required because Kubernetes combines lists which is often not the desired use
    """
    # If type mismatch then assume replace
    if (
        desired_manifest
        and current_manifest
        and type(desired_manifest) is not type(current_manifest)
    ):
        log.debug4("Requires replace due to type mismatch")
        return True

    if isinstance(current_manifest, list) and isinstance(desired_manifest, list):
        # if the desired has less then the current then return True. Removing
        # from a list requires a put
        if len(current_manifest) > len(desired_manifest):
            log.debug4("Requires replace due to list deletion")
            return True
        # Iterate over the desired manifest
        for recurse_a, recurse_b in zip(current_manifest, desired_manifest):
            if modified_lists(recurse_a, recurse_b, in_list=True):
                return True
    if isinstance(current_manifest, dict) and isinstance(desired_manifest, dict):
        key_intersection = set(current_manifest.keys()).intersection(
            set(desired_manifest.keys())
        )
        # If there are no common keys and we're in a list then return True
        # this means we have a new object
        if in_list and not key_intersection:
            log.debug4("Requires replace due to no common key in list")
            return True

        # Tack if one key has the same value. This is
        # required for kubernetes merges
        at_least_one_common = False
        for k in key_intersection:
            # Check if two objects are the same for their value operations
            changed = False
            if isinstance(current_manifest[k], list):
                changed = bool(
                    recursive_list_diff(current_manifest[k], desired_manifest[k])
                )
            elif isinstance(current_manifest[k], dict):
                changed = bool(recursive_diff(current_manifest[k], desired_manifest[k]))
            else:
                changed = current_manifest[k] != desired_manifest[k]

            if not changed:
                at_least_one_common = True

            # Recurse on modified lists
            if modified_lists(current_manifest[k], desired_manifest[k]):
                return True
        if in_list and not at_least_one_common:
            log.debug4("Requires replace due to no common key/value in list")
            return True
    return False


def modified_value_from(current_manifest: Any, desired_manifest: Any) -> bool:
    """Helper function to check if a manifest switched from value to valueFrom. These are mutually
    exclusive thus they require a replace command.
    """
    # If type mismatch then assume replace
    if (
        desired_manifest
        and current_manifest
        and type(desired_manifest) is not type(current_manifest)
    ):
        log.debug4("Requires replace due to type mismatch")
        return True

    if isinstance(current_manifest, list) and isinstance(desired_manifest, list):
        for recurse_a, recurse_b in zip(current_manifest, desired_manifest):
            if modified_value_from(recurse_a, recurse_b):
                return True
    if isinstance(current_manifest, dict) and isinstance(desired_manifest, dict):
        if ("value" in current_manifest and "valueFrom" in desired_manifest) or (
            "valueFrom" in current_manifest and "value" in desired_manifest
        ):
            log.debug4("Requires replace due to value/valueFrom change")
            return True
        else:
            for k in set(current_manifest.keys()).intersection(
                set(desired_manifest.keys())
            ):
                if modified_value_from(current_manifest[k], desired_manifest[k]):
                    return True
    return False


_REPLACE_FUNCS: List[Callable[[str, str], bool]] = [modified_lists, modified_value_from]


def requires_replace(current_manifest: dict, desired_manifest: dict) -> bool:
    """Function to determine if a resource requires a replace operation instead
    of apply. This can occur due to list merging, or updating envVars

    Args:
        current_manifest (dict): The current manifest in the cluster
        desired_manifest (dict): The desired manifest that should be applied

    Returns:
        bool: If the current manifest requires a replace operation
    """
    for func in _REPLACE_FUNCS:
        if func(current_manifest, desired_manifest):
            log.debug4("Manifest requires replace", desired_manifest)
            return True
    return False



"""
Test the construction and management of status objects
"""

# Standard
from datetime import datetime
import copy

# Third Party
import pytest

# First Party
import alog

# Local
from oper8 import status
from oper8.dag import CompletionState, Node
from oper8.status import ReadyReason, ServiceStatus, UpdatingReason
from oper8.test_helpers.helpers import configure_logging, library_config
from oper8.utils import nested_get

configure_logging()

log = alog.use_channel("TEST")

## Helpers #####################################################################


def strip_timestamps(res):
    for entry in res.get("conditions", []):
        if entry["type"] in [status.READY_CONDITION, status.UPDATING_CONDITION]:
            assert "lastTransactionTime" in entry
        if "lastTransactionTime" in entry:
            del entry["lastTransactionTime"]
    return res


## make_application_status #####################################################


def test_make_application_status_empty():
    """Make sure only specified status elements are present in the output"""
    res = status.make_application_status()
    assert res == {"conditions": []}


def test_make_application_status_partial():
    """Make sure a subset of the conditions can be given"""
    res = status.make_application_status(ready_reason=status.ReadyReason.STABLE)
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.STABLE.value,
                "status": "True",
                "message": "",
            }
        ]
    }


def test_make_application_status_reason_all_args():
    """Make sure that all args make it through to the output"""
    version = "1.0.0"
    res = status.make_application_status(
        ready_reason=status.ReadyReason.INITIALIZING,
        ready_message="Ready Message",
        updating_reason=status.UpdatingReason.PRECONDITION_WAIT,
        updating_message="Updating Message",
        version=version,
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.INITIALIZING.value,
                "status": "False",
                "message": "Ready Message",
            },
            {
                "type": status.UPDATING_CONDITION,
                "reason": status.UpdatingReason.PRECONDITION_WAIT.value,
                "status": "True",
                "message": "Updating Message",
            },
        ],
        "versions": {"reconciled": version},
    }


def test_make_application_status_reason_str():
    """Make sure that a reason argument can be given as a string"""
    res = status.make_application_status(
        ready_reason=status.ReadyReason.INITIALIZING.value,
        updating_reason=status.UpdatingReason.PRECONDITION_WAIT.value,
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.INITIALIZING.value,
                "status": "False",
                "message": "",
            },
            {
                "type": status.UPDATING_CONDITION,
                "reason": status.UpdatingReason.PRECONDITION_WAIT.value,
                "status": "True",
                "message": "",
            },
        ]
    }


def test_make_application_status_invalid_reason_enum():
    """Make sure that an invalid reason enum value throws an appropriate error"""
    with pytest.raises(ValueError):
        status.make_application_status(ready_reason="foobar")


def test_make_application_status_component_state():
    """Make sure that component_state gets correctly incorporated into status"""
    res = status.make_application_status(
        component_state=CompletionState(
            verified_nodes=[Node("B"), Node("A")],
            unverified_nodes=[Node("C")],
            failed_nodes=[Node("D")],
            unstarted_nodes=[Node("E")],
        )
    )
    assert res == {
        "conditions": [],
        status.COMPONENT_STATUS: {
            status.COMPONENT_STATUS_ALL_NODES: ["A", "B", "C", "D", "E"],
            status.COMPONENT_STATUS_DEPLOYED_NODES: ["A", "B", "C"],
            status.COMPONENT_STATUS_UNVERIFIED_NODES: ["C"],
            status.COMPONENT_STATUS_FAILED_NODES: ["D"],
            status.COMPONENT_STATUS_DEPLOYED_COUNT: "3/5",
            status.COMPONENT_STATUS_VERIFIED_COUNT: "2/5",
        },
    }


def test_make_application_status_supported_versions():
    """Make sure that setting supported versions in status creates the correct
    status.versions entries
    """
    versions = ["one", "two", "3", "v4.0.0"]
    res = status.make_application_status(supported_versions=versions)
    assert nested_get(res, status.VERSIONS_FIELD_AVAILABLE_VERSIONS) == [
        {"name": version} for version in versions
    ]


@pytest.mark.parametrize(
    ["ready_reason", "updating_reason", "service_status"],
    [
        [
            ReadyReason.STABLE,
            UpdatingReason.STABLE,
            ServiceStatus.COMPLETED,
        ],
        [
            ReadyReason.STABLE.value,
            UpdatingReason.STABLE.value,
            ServiceStatus.COMPLETED,
        ],  # make sure string type reason value is handled correctly
        [
            ReadyReason.INITIALIZING,
            UpdatingReason.VERIFY_WAIT,
            ServiceStatus.IN_PROGRESS,
        ],  # waiting install completion scenario
        [
            ReadyReason.STABLE,
            UpdatingReason.VERSION_CHANGE,
            ServiceStatus.IN_PROGRESS,
        ],  # upgrade scenario
        [
            ReadyReason.INITIALIZING,
            UpdatingReason.ERRORED,
            ServiceStatus.FAILED,
        ],  # error scenario
        [
            ReadyReason.ERRORED,
            UpdatingReason.VERIFY_WAIT,
            ServiceStatus.FAILED,
        ],  # error scenario
        [
            None,
            None,
            ServiceStatus.IN_PROGRESS,
        ],  # starting reconciliation
    ],
)
def test_make_application_status_service_status(
    ready_reason, updating_reason, service_status
):
    """Make sure service status is correctly set based on current ready/updating status"""
    kind = "SampleService"
    res = status.make_application_status(
        ready_reason=ready_reason, updating_reason=updating_reason, kind=kind
    )

    assert res.get("sampleServiceStatus", "") == service_status.value


def test_make_application_status_operator_version():
    """Make sure that setting operator version in status creates the correct
    status.versions entry
    """
    operator_version = "1.2.3"
    res = status.make_application_status(operator_version=operator_version)
    assert nested_get(res, status.OPERATOR_VERSION) == "1.2.3"


def test_make_application_status_no_operator_version():
    """Make sure that None operator version does not error"""
    operator_version = None
    res = status.make_application_status(operator_version=operator_version)


## update_application_status ###################################################


def test_update_application_status_empty_current():
    """Make sure an empty current_status dict is handled by using the defaults
    from make_application_status
    """
    res = status.update_application_status(
        current_status={},
        ready_reason=status.ReadyReason.STABLE,
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.STABLE.value,
                "status": "True",
                "message": "",
            }
        ]
    }


def test_update_application_status_no_current_override():
    """Make sure a value in current_status without an override ends up in the
    output status
    """
    res = status.update_application_status(
        current_status=status.make_application_status(
            ready_reason=status.ReadyReason.STABLE, version="1.0.0"
        )
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.STABLE.value,
                "status": "True",
                "message": "",
            }
        ],
        "versions": {"reconciled": "1.0.0"},
    }


def test_update_application_status_with_current_override():
    """Make sure a value in current_status and an override kwarg uses the
    override
    """
    res = status.update_application_status(
        current_status=status.make_application_status(
            ready_reason=status.ReadyReason.STABLE
        ),
        ready_reason=status.ReadyReason.ERRORED,
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.ERRORED.value,
                "status": "False",
                "message": "",
            }
        ]
    }


def test_update_application_status_with_all_current():
    """Make sure that when the status has all entries, they all pass through or
    are overridden as expected
    """
    res = status.update_application_status(
        current_status=status.make_application_status(
            ready_reason=status.ReadyReason.STABLE,
            updating_reason=status.UpdatingReason.PRECONDITION_WAIT,
            version="1.0.0",
        ),
        ready_reason=status.ReadyReason.ERRORED,
        updating_reason=status.UpdatingReason.ERRORED,
        version="2.0.0",
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.ERRORED.value,
                "status": "False",
                "message": "",
            },
            {
                "type": status.UPDATING_CONDITION,
                "reason": status.UpdatingReason.ERRORED.value,
                "status": "False",
                "message": "",
            },
        ],
        "versions": {
            "reconciled": "2.0.0",
        },
    }


def test_update_application_status_external_conditions():
    """Make sure that conditions managed by other parts of the operator are
    preserved when updating status.
    """
    external_condition = {"type": "Foo", "bar": "baz"}
    res = status.update_application_status(
        current_status={"conditions": [external_condition]},
        ready_reason=status.ReadyReason.ERRORED,
    )
    assert strip_timestamps(res) == {
        "conditions": [
            {
                "type": status.READY_CONDITION,
                "reason": status.ReadyReason.ERRORED.value,
                "status": "False",
                "message": "",
            },
            external_condition,
        ]
    }


def test_update_application_status_external_status():
    """Make sure that non-condition status elements managed by other parts of
    the operator are preserved when updating status.
    """
    external_status = {"foo": "bar", "baz": "bat"}
    res = status.update_application_status(
        current_status=external_status,
        ready_reason=status.ReadyReason.ERRORED,
    )
    expected = copy.copy(external_status)
    expected["conditions"] = [
        {
            "type": status.READY_CONDITION,
            "reason": status.ReadyReason.ERRORED.value,
            "status": "False",
            "message": "",
        },
    ]

    assert strip_timestamps(res) == expected


def test_update_application_status_invalid_kwarg():
    """Make sure an invalid kwarg throws an error"""
    with pytest.raises(TypeError):
        res = status.update_application_status(current_status={}, foo="bar")


def test_update_application_status_no_mutation():
    """Make sure that the original status object is not mutated"""
    current_status = status.make_application_status(
        ready_reason=status.ReadyReason.STABLE,
        version="old version",
    )
    current_status_copy = copy.deepcopy(current_status)
    status.update_application_status(
        current_status=current_status,
        ready_reason=status.ReadyReason.ERRORED,
        version="new version",
    )
    assert current_status == current_status_copy


def test_update_application_status_supported_versions_not_changed():
    """Make sure that updating a status object which contains supported versions
    does not remove them
    """
    versions = ["one", "two", "3", "v4.0.0"]
    initial_status = status.make_application_status(supported_versions=versions)
    updated_status = status.update_application_status(
        current_status=initial_status, ready_reason=status.ReadyReason.STABLE
    )
    assert nested_get(updated_status, status.VERSIONS_FIELD_AVAILABLE_VERSIONS) == [
        {"name": version} for version in versions
    ]


def test_update_application_status_operator_version():
    """Make sure that the non-null config.operator_version is pulled correctly"""
    with library_config(operator_version="1.2.3"):
        updated_cfg = status.update_application_status({})
    assert nested_get(updated_cfg, status.OPERATOR_VERSION) == "1.2.3"


def test_update_application_status_no_operator_version():
    """Make sure that the null config.operator_version and null status.version does not error or display"""
    with library_config(operator_version=None):
        updated_cfg = status.update_application_status({})
    assert status.OPERATOR_VERSION.split(".")[0] not in updated_cfg


def test_update_application_status_version_no_operator_version():
    """Make sure that the null config.operator_version does not error or display in non-null status.version"""
    with library_config(operator_version=None):
        updated_cfg = status.update_application_status(
            current_status={}, version="1.2.3"
        )
    assert status.OPERATOR_VERSION not in updated_cfg


## status_changed ##############################################################


def test_status_changed_different_types():
    """Make sure that if the statuses are different types, a change is indicated"""
    assert status.status_changed(1, "two")
    assert status.status_changed(None, {})


def test_status_changed_different_non_dicts():
    """Make sure that if the statuses are not dicts, a change is indicated"""
    assert status.status_changed("one", "one")


def test_status_changed_no_conditions():
    """Make sure that without conditions, a deep-equal is used"""
    ts_a = datetime.now().isoformat()
    ts_b = datetime.now().isoformat()
    assert not status.status_changed({"a": 1, "b": ts_a}, {"a": 1, "b": ts_a})
    assert status.status_changed({"a": 1, "b": ts_a}, {"a": 1, "b": ts_b})


def test_status_changed_no_conditions_cur():
    """Make sure that with conditions in the new status, but not in the current,
    a change is indicated
    """
    assert status.status_changed({}, {"conditions": []})


def test_status_changed_conditions_different_len():
    """Make sure that with differnt length condition lists, a change is
    indicated
    """
    assert status.status_changed(
        {"conditions": [{"type": "a"}]},
        {"conditions": [{"type": "a"}, {"type": "b"}]},
    )


def test_status_changed_condition_type_mismatch():
    """Make sure that if the elements of a conditions list have different types
    change is indicated
    """
    assert status.status_changed(
        {"conditions": [None]},
        {"conditions": [{"type": "a"}]},
    )


def test_status_changed_conditions_match_different_timestamps():
    """Make sure that if the conditions match but have different timestamps, a
    change is not indicated
    """
    current_status = status.make_application_status(
        ready_reason=status.ReadyReason.STABLE
    )
    new_status = status.make_application_status(ready_reason=status.ReadyReason.STABLE)
    assert current_status != new_status
    assert not status.status_changed(current_status, new_status)


def test_status_changed_conditions_mismatch_different_values():
    """Make sure that if the conditions have different values, a change is
    indicated
    """
    current_status = status.make_application_status(
        ready_reason=status.ReadyReason.STABLE
    )
    new_status = status.make_application_status(ready_reason=status.ReadyReason.ERRORED)
    assert status.status_changed(current_status, new_status)


def test_status_changed_non_list_conditions_error():
    """Make sure that if conditions are not a list, we raise an error"""
    with pytest.raises(AssertionError):
        assert status.status_changed(
            {"conditions": {}},
            {"conditions": {}},
        )


def test_status_changed_condition_type_non_dict():
    """Make sure that if the elements of a conditions list are not dicts, an
    error is thrown
    """
    with pytest.raises(AssertionError):
        assert status.status_changed(
            {"conditions": [1]},
            {"conditions": [1]},
        )


## get_condition ###############################################################


def test_get_condition_when_present():
    """Make sure get_condition returns a condition when it is found"""
    st = status.make_application_status(ready_reason=status.ReadyReason.STABLE)
    assert status.get_condition(status.READY_CONDITION, strip_timestamps(st)) == {
        "type": status.READY_CONDITION,
        "reason": status.ReadyReason.STABLE.value,
        "status": "True",
        "message": "",
    }


def test_get_condition_when_missing():
    """Make sure get_condition returns an empty dict when the requested
    condition is missing
    """
    st = status.make_application_status(ready_reason=status.ReadyReason.STABLE)
    assert status.get_condition(status.UPDATING_CONDITION, st) == {}


def test_get_condition_no_conditions():
    """Make sure get_condition returns an empty dict when the current status is
    missing the conditions section
    """
    assert status.get_condition(status.UPDATING_CONDITION, {"foo": "bar"}) == {}


## get_version ###############################################################


@pytest.mark.parametrize(["version"], [["1.0.0"], [None]])
def test_get_version(version):
    """Make sure current verified version can be obtained"""
    st = status.make_application_status(version=version)
    assert version == status.get_version(st)


## Status Mappings #############################################################


def check_status_map(condition_fn, status_map):
    for reason, status in status_map.items():
        cond = condition_fn(
            reason=reason, message="", last_transaction_time=datetime.now()
        )
        assert str(status) == cond["status"]


def test_ready_status_map():
    """Make sure the status bools are set correctly for each ReadyReason"""
    check_status_map(
        status._make_ready_condition,
        {
            status.ReadyReason.INITIALIZING: False,
            status.ReadyReason.STABLE: True,
            status.ReadyReason.CONFIG_ERROR: False,
            status.ReadyReason.ERRORED: False,
        },
    )


def test_updating_status_map():
    """Make sure the status bools are set correctly for each UpdatingReason"""
    check_status_map(
        status._make_updating_condition,
        {
            status.UpdatingReason.STABLE: False,
            status.UpdatingReason.PRECONDITION_WAIT: True,
            status.UpdatingReason.VERIFY_WAIT: True,
            status.UpdatingReason.CLUSTER_ERROR: False,
            status.UpdatingReason.ERRORED: False,
        },
    )



"""
Base class for all Datastore component implementations
"""

# Standard
from typing import Optional
import abc

# First Party
import aconfig

# Local
from ... import Session
from ..oper8x_component import Oper8xComponent
from .connection_base import DatastoreConnectionBase


class Datastore(Oper8xComponent):
    """
    The Datastore baseclass defines the interface that any datastore must
    conform to. It is a oper8 Component and should be constructed via a per-type
    factory.
    """

    _TYPE_LABEL_ATTRIBUTE = "TYPE_LABEL"

    def __init__(
        self,
        session: Session,
        config: aconfig.Config,
        instance_name: Optional[str] = None,
        disabled: bool = False,
    ):
        """This passthrough constructor enforces that all datastores have a
        class attribute TYPE_LABEL (str)
        """
        type_label = getattr(self, self._TYPE_LABEL_ATTRIBUTE, None)
        assert isinstance(
            type_label, str
        ), f"All datastores types must define {self._TYPE_LABEL_ATTRIBUTE} as a str"
        super().__init__(session=session, disabled=disabled)
        self._config = config
        self.instance_name = instance_name

    @property
    def config(self) -> aconfig.Config:
        """The config for this instance of the datastore"""
        return self._config

    @abc.abstractmethod
    def get_connection(self) -> DatastoreConnectionBase:
        """Get the connection object for this datastore instance. Each datastore
        type must manage a common abstraction for a connection which clients
        will use to connect to the datastore.
        """



"""Tests for the Controller class"""

# Standard
from datetime import timedelta
from functools import partial

# Third Party
import pytest

# First Party
import aconfig
import alog

# Local
from oper8 import Component, Controller, ReconcileManager, status
from oper8.dag import Node
from oper8.exceptions import PreconditionError, RolloutError, VerificationError
from oper8.reconcile import RequeueParams
from oper8.test_helpers.helpers import (
    TEST_NAMESPACE,
    DummyController,
    DummyNodeComponent,
    FailOnce,
    MockDeployManager,
    library_config,
    setup_cr,
    setup_session,
)
from tests.test_reconcile import ReconcileDummyController

log = alog.use_channel("TEST")

################################################################################
## Helpers #####################################################################
################################################################################


class AlogConfigureMock:
    def __init__(self):
        self.kwargs = None

    def __call__(self, **kwargs):
        self.kwargs = kwargs


def check_status(deploy_manager, cr, ready_reason=None, updating_reason=None):
    """Shared helper for checking status after reconcile"""
    obj = deploy_manager.get_obj(
        kind=cr.kind,
        name=cr.metadata.name,
        namespace=cr.metadata.namespace,
        api_version=cr.apiVersion,
    )
    assert obj is not None

    ready_cond = status.get_condition(status.READY_CONDITION, obj.status)
    if ready_reason:
        assert ready_cond
        assert ready_cond["reason"] == ready_reason.value
    else:
        assert not ready_cond

    update_cond = status.get_condition(status.UPDATING_CONDITION, obj.status)
    if updating_reason:
        assert update_cond
        assert update_cond["reason"] == updating_reason.value
    else:
        assert not update_cond


def check_bad_reconcile(
    controller_class,
    ready_reason=None,
    updating_reason=None,
    raises=True,
):
    """Common test body for tests of reconcile that raise unexpected errors"""
    cr = setup_cr()
    dm = MockDeployManager(resources=[cr])
    ctrlr = controller_class(deploy_manager=dm)

    # Make sure RolloutError is raised
    if raises:
        with pytest.raises(RolloutError):
            ctrlr.reconcile(cr)
    else:
        result = ctrlr.reconcile(cr)
        assert result.requeue

    # Make sure status is set correctly
    check_status(dm, cr, ready_reason, updating_reason)


class FinalizerController(Controller):
    group = "foo.bar"
    version = "v1"
    kind = "Foo"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setup_called = False
        self.finalize_called = False

    def setup_components(self, session):
        self.setup_called = True

    def finalize_components(self, session):
        self.finalize_called = True


################################################################################
## Tests #######################################################################
################################################################################

##################
## Construction ##
##################


def test_construct_defaults():
    """Make sure that a controller can be constructed with its default args"""
    with library_config(dry_run=True):
        DummyController()
    with library_config(dry_run=False):
        DummyController()


def test_construct_input_args():
    """Make sure that a controller can be constructed with given args"""
    DummyController(config_defaults=aconfig.Config({"foo": 1}))


def test_construct_property_check():
    """Make sure that a controller can not be constructed without the required
    abstract class properties defined
    """

    class BadControllerNoGroup(Controller):
        version = "v1"
        kind = "Foo"

        def setup_components(*_, **__):
            pass

    class BadControllerNoVersion(Controller):
        group = "foo.bar"
        kind = "Foo"

        def setup_components(*_, **__):
            pass

    class BadControllerNoKind(Controller):
        group = "foo.bar"
        version = "v1"

        def setup_components(*_, **__):
            pass

    with pytest.raises(NotImplementedError):
        BadControllerNoGroup()

    with pytest.raises(NotImplementedError):
        BadControllerNoVersion()

    with pytest.raises(NotImplementedError):
        BadControllerNoKind()


#######################
## defaults ##
#######################


def test_get_cr_manifest_defaults():
    class TestDummyController(DummyController):
        def get_cr_manifest_defaults(
            self,
        ):
            return aconfig.Config({"example": "config"})

    ctrlr = TestDummyController()
    assert ctrlr.get_cr_manifest_defaults() == aconfig.Config({"example": "config"})


def test_get_config_defaults():
    class TestDummyController(DummyController):
        def get_config_defaults(
            self,
        ):
            return aconfig.Config({"example": "config"})

    ctrlr = TestDummyController()
    assert ctrlr.get_config_defaults() == aconfig.Config({"example": "config"})

    ctrlr = DummyController(config_defaults={"example": "anotherconfig"})
    assert ctrlr.get_config_defaults() == aconfig.Config({"example": "anotherconfig"})


#######################
## _manage_components ##
#######################


def test_manage_components():
    """Make sure manage_components correctly calls either setup_components
    or finalize_components
    """
    session = setup_session()
    ctrlr = ReconcileDummyController()
    ctrlr._manage_components(session, is_finalizer=False)
    assert ctrlr.setup_components.called
    assert not ctrlr.finalize_components.called
    assert not session.graph.empty()
    assert Node("foo") in session.graph


def test_manage_components_finalizer():
    """Make sure manage_components correctly calls either setup_components
    or finalize_components
    """
    session = setup_session()
    ctrlr = ReconcileDummyController()
    ctrlr._manage_components(session, is_finalizer=True)
    assert not ctrlr.setup_components.called
    assert ctrlr.finalize_components.called
    assert session.graph.empty()


#######################
## _rollout_components ##
#######################


def test_rollout_components():
    session = setup_session()
    ctrlr = ReconcileDummyController()

    # Deploy initial component object to ensure disabled
    # components get removed
    session.deploy_manager.deploy(
        [
            {
                "kind": "Baz",
                "apiVersion": "v3",
                "metadata": {"name": "baz", "namespace": session.namespace},
            }
        ]
    )
    assert session.deploy_manager.has_obj(
        kind="Baz", api_version="v3", name="baz", namespace=session.namespace
    )

    ctrlr.setup_components(session)
    completion_state = ctrlr._rollout_components(session)

    # Ensure the cluster state is as expected
    assert session.deploy_manager.has_obj(
        kind="Foo", api_version="v1", name="foo", namespace=session.namespace
    )
    assert session.deploy_manager.has_obj(
        kind="Bar", api_version="v2", name="bar", namespace=session.namespace
    )
    assert not session.deploy_manager.has_obj(
        kind="Baz", api_version="v3", name="baz", namespace=session.namespace
    )

    # Make sure the after_deploy and after_verify were called
    assert ctrlr.after_deploy.called
    assert not ctrlr.after_deploy_unsuccessful.called
    assert ctrlr.after_verify.called
    assert not ctrlr.after_verify_unsuccessful.called

    # Test completion state
    assert completion_state.deploy_completed()
    assert completion_state.verify_completed()
    assert not completion_state.failed()
    assert Node("foo") in completion_state.verified_nodes
    assert Node("bar") in completion_state.verified_nodes
    assert Node("baz") in completion_state.verified_nodes


@pytest.mark.parametrize(
    ["raised_exception", "expected_exception"],
    [
        [ValueError, RolloutError],
        [PreconditionError, PreconditionError],
    ],
)
def test_rollout_components_exception(raised_exception, expected_exception):
    session = setup_session()

    ctrlr = DummyController(
        components=[{"name": "bad", "deploy_fail": raised_exception}]
    )
    ctrlr.setup_components(session)
    with pytest.raises(expected_exception):
        ctrlr._rollout_components(session)


###############
## reconcile ##
###############


def test_run_reconcile_ok():
    """Test that a reconciliation with several components that complete cleanly
    exists correctly and sets status conditions to STABLE
    """
    session = setup_session()
    ctrlr = DummyController(
        components=[
            {
                "name": "foo",
                "api_objects": [("foo", {"kind": "Foo", "apiVersion": "v1"})],
            },
            {
                "name": "bar",
                "api_objects": [("bar", {"kind": "Bar", "apiVersion": "v2"})],
                "upstreams": ["foo"],
            },
        ],
    )
    completion_state = ctrlr.run_reconcile(session)
    assert session.deploy_manager.has_obj(
        kind="Foo", name="foo", api_version="v1", namespace=session.namespace
    )
    assert session.deploy_manager.has_obj(
        kind="Bar", name="bar", api_version="v2", namespace=session.namespace
    )

    assert completion_state.deploy_completed()
    assert completion_state.verify_completed()
    assert not completion_state.failed()
    assert Node("foo") in completion_state.verified_nodes
    assert Node("bar") in completion_state.verified_nodes


def test_run_reconcile_rerun():
    """Test that when a rollout is incomplete at the deploy phase, the
    PreconditionError is correctly propagated and the status is set correctly
    """
    session = setup_session()
    ctrlr = ReconcileDummyController()
    ctrlr.run_reconcile(session)

    with pytest.raises(RolloutError):
        ctrlr.run_reconcile(session)


##################
## after_deploy ##
##################


def test_after_deploy_failure():
    """Make sure that a raised error during after_deploy is handled and the
    status is set correctly
    """
    session = setup_session()
    ctrlr = DummyController(after_deploy_fail=True)

    with pytest.raises(VerificationError):
        ctrlr.run_reconcile(session)

    assert ctrlr.after_deploy.called
    assert not ctrlr.after_deploy_unsuccessful.called
    assert not ctrlr.after_verify.called
    assert not ctrlr.after_verify_unsuccessful.called


def test_after_deploy_error():
    """Make sure that an error in after_deploy is handled"""
    session = setup_session()
    ctrlr = DummyController(after_deploy_fail="assert")
    with pytest.raises(RolloutError):
        ctrlr.run_reconcile(session)

    # Make sure after_deploy was called, but after_verify was not
    assert ctrlr.after_deploy.called
    assert not ctrlr.after_deploy_unsuccessful.called
    assert not ctrlr.after_verify.called
    assert not ctrlr.after_verify_unsuccessful.called


##################
## after_verify ##
##################


def test_after_verify_failure():
    """Make sure that a negative return from after_verify is handled and the
    status is set correctly
    """
    session = setup_session()
    ctrlr = DummyController(after_verify_fail=True)
    with pytest.raises(VerificationError):
        ctrlr.run_reconcile(session)

    # Make sure both after_deploy and after_verify were called
    assert ctrlr.after_deploy.called
    assert not ctrlr.after_deploy_unsuccessful.called
    assert ctrlr.after_verify.called
    assert not ctrlr.after_verify_unsuccessful.called


def test_after_verify_error():
    """Make sure that an error in after_verify is handled"""
    session = setup_session()
    ctrlr = DummyController(after_verify_fail="assert")
    with pytest.raises(RolloutError):
        ctrlr.run_reconcile(session)

    # Make sure both after_deploy and after_verify were called
    assert ctrlr.after_deploy.called
    assert not ctrlr.after_deploy_unsuccessful.called
    assert ctrlr.after_verify.called
    assert not ctrlr.after_verify_unsuccessful.called


####################
## should_requeue ##
####################


@pytest.mark.parametrize(
    ["requeue", "requeue_params"],
    [
        [True, None],
        [False, None],
        [True, RequeueParams(requeue_after=timedelta(seconds=300.0))],
    ],
)
def test_override_should_requeue(requeue, requeue_params):
    """Make sure overriding should_requeue by children works"""

    def should_requeue(*_, **__):
        return requeue, requeue_params

    class CustomRequeueController(DummyController):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.should_requeue = should_requeue

    session = setup_session()
    ctrlr = CustomRequeueController(
        components=[],
    )
    ctrlr.run_reconcile(session)
    returned_requeue, returned_params = ctrlr.should_requeue()

    assert returned_requeue == requeue
    assert returned_params == requeue_params


def test_should_requeue_error():
    """Make sure status becomes error when should_requeue raises unexpected error"""

    def should_requeue(*_, **__):
        raise RuntimeError("Oh No!")

    class CustomRequeueController(DummyController):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.should_requeue = should_requeue

    session = setup_session()
    ctrlr = CustomRequeueController()

    with pytest.raises(RuntimeError):
        ctrlr.should_requeue(session)


def test_should_requeue_no_obj():
    session = setup_session(deploy_initial_cr=False)
    ctrlr = DummyController()
    assert ctrlr.should_requeue(session) == (False, RequeueParams())


def test_should_requeue_failed_status():
    """Make sure status becomes error when should_requeue raises unexpected error"""
    dm = MockDeployManager(get_state_fail=FailOnce((False, None), fail_number=2))
    session = setup_session(deploy_manager=dm)
    ctrlr = DummyController()

    assert ctrlr.should_requeue(session) == (True, RequeueParams())


###########
## other ##
###########


def test_str():
    """Make sure the group/version/kind are present in the str representation"""
    ctrlr = DummyController()
    assert ctrlr.group in str(ctrlr)
    assert ctrlr.version in str(ctrlr)
    assert ctrlr.kind in str(ctrlr)


def test_has_finalizer():
    """Make sure that the has_finalizer property only returns true when a valid
    finalize_components implementation is provided
    """

    class WithFinalizer(Controller):
        group = "foo.bar"
        version = "v1"
        kind = "Foo"

        def setup_components(self, session):
            pass

        def finalize_components(self, session):
            pass

    class WithoutFinalizer(Controller):
        group = "foo.bar"
        version = "v2"
        kind = "Bar"

        def setup_components(self, session):
            pass

    assert WithFinalizer.has_finalizer
    assert not WithoutFinalizer.has_finalizer


class FooComponent(DummyNodeComponent):
    name = "foo"

    def __init__(self, session):
        super().__init__(
            session=session,
            api_objects=[("foo", {"kind": "Foo", "apiVersion": "v1"})],
        )


class SubsystemController(Controller):
    group = "foo.bar"
    version = "v1"
    kind = "Subsystem"

    def setup_components(self, session):
        FooComponent(session)


class SubsystemComponent(Component):
    name = "Subsystem"

    def build_chart(self, session):
        self.add_resource(
            "subsystem",
            {
                "kind": "Subsystem",
                "apiVersion": "foo.bar/v1",
                "metadata": {"name": "subsystem", "namespace": TEST_NAMESPACE},
            },
        )


class BarComponent(DummyNodeComponent):
    name = "bar"

    def __init__(self, session):
        super().__init__(
            session=session,
            api_objects=[("bar", {"kind": "Bar", "apiVersion": "v1"})],
        )


class ParentController(Controller):
    group = "foo.bar"
    version = "v1"
    kind = "Parent"

    def setup_components(self, session):
        BarComponent(session)
        SubsystemComponent(session)


def test_subsystem_rollout():
    """Test that multiple controllers can work together to rollout a tree of
    components and subsystems
    """

    # Set up the CR for the parent controller
    cr = setup_cr(api_version="foo.bar/v1", kind="Parent", version="v1")

    # Instantiate the controllers with a shared deploy manager
    dm = MockDeployManager()
    rm = ReconcileManager(deploy_manager=dm, reimport_controller=False)
    parent_ctrlr = ParentController()
    parent_reconcile = partial(rm.safe_reconcile, parent_ctrlr)
    subsystem_ctrlr = SubsystemController()
    susbsystem_reconcile = partial(rm.safe_reconcile, subsystem_ctrlr)

    # Register both Controllers to watch their respective resources
    dm.register_watch(
        api_version="foo.bar/v1",
        kind="Parent",
        callback=parent_reconcile,
    )
    dm.register_watch(
        api_version="foo.bar/v1",
        kind="Subsystem",
        callback=susbsystem_reconcile,
    )

    # Deploy the parent CR
    dm.deploy([cr])

    # Make sure the api objects from both controllers were deployed
    assert dm.has_obj(
        kind="Foo", name="foo", api_version="v1", namespace=cr.metadata.namespace
    )
    assert dm.has_obj(
        kind="Bar", name="bar", api_version="v1", namespace=cr.metadata.namespace
    )
